//
//  PithosActivityFacility.h
//  pithos-macos
//
// Copyright 2011-2012 GRNET S.A. All rights reserved.
//
// Redistribution and use in source and binary forms, with or
// without modification, are permitted provided that the following
// conditions are met:
// 
//   1. Redistributions of source code must retain the above
//      copyright notice, this list of conditions and the following
//      disclaimer.
// 
//   2. Redistributions in binary form must reproduce the above
//      copyright notice, this list of conditions and the following
//      disclaimer in the documentation and/or other materials
//      provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
// OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
// USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
// AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and
// documentation are those of the authors and should not be
// interpreted as representing official policies, either expressed
// or implied, of GRNET S.A.

#import <Foundation/Foundation.h>
#import "PithosActivity.h"
@class PithosAccount;

@protocol PithosActivityFacilityDelegate <NSObject>
@required
- (void)activityUpdate:(NSDictionary *)info;
@end

@interface PithosActivityFacility : NSObject {
    NSMutableArray *runningActivities;
    NSMutableArray *endingActivities;
    NSUInteger totalUploadBytes;
    NSUInteger currentUploadBytes;
    NSUInteger totalDownloadBytes;
    NSUInteger currentDownloadBytes;

    NSTimer *timer;
    BOOL pickedRunning;
    
    id<PithosActivityFacilityDelegate> __unsafe_unretained delegate;
}

@property (nonatomic, unsafe_unretained) id delegate;

+ (id)defaultPithosActivityFacility;

- (void)reset;

- (PithosActivity *)startActivityWithType:(PithosActivityType)type 
                                  message:(NSString *)message 
                               totalBytes:(NSUInteger)totalBytes 
                             currentBytes:(NSUInteger)currentBytes 
                            pithosAccount:(PithosAccount *)pithosAccount;
- (PithosActivity *)startActivityWithType:(PithosActivityType)type 
                                  message:(NSString *)message 
                               totalBytes:(NSUInteger)totalBytes 
                             currentBytes:(NSUInteger)currentBytes;
- (PithosActivity *)startActivityWithType:(PithosActivityType)type message:(NSString *)message pithosAccount:(PithosAccount *)pithosAccount;
- (PithosActivity *)startActivityWithType:(PithosActivityType)type message:(NSString *)message;
- (PithosActivity *)startAndEndActivityWithType:(PithosActivityType)type message:(NSString *)message pithosAccount:(PithosAccount *)pithosAccount;
- (PithosActivity *)startAndEndActivityWithType:(PithosActivityType)type message:(NSString *)message;

- (void)updateActivity:(PithosActivity *)activity 
           withMessage:(NSString *)message 
            totalBytes:(NSUInteger)totalBytes 
          currentBytes:(NSUInteger)currentBytes;
- (void)updateActivity:(PithosActivity *)activity 
           withMessage:(NSString *)message;
- (void)endActivity:(PithosActivity *)activity 
        withMessage:(NSString *)message 
         totalBytes:(NSUInteger)totalBytes 
       currentBytes:(NSUInteger)currentBytes;
- (void)endActivity:(PithosActivity *)activity 
        withMessage:(NSString *)message;

@end