//
//  PithosUtilities.h
//  pithos-macos
//
// Copyright 2011-2012 GRNET S.A. All rights reserved.
//
// Redistribution and use in source and binary forms, with or
// without modification, are permitted provided that the following
// conditions are met:
// 
//   1. Redistributions of source code must retain the above
//      copyright notice, this list of conditions and the following
//      disclaimer.
// 
//   2. Redistributions in binary form must reproduce the above
//      copyright notice, this list of conditions and the following
//      disclaimer in the documentation and/or other materials
//      provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
// OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
// USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
// AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and
// documentation are those of the authors and should not be
// interpreted as representing official policies, either expressed
// or implied, of GRNET S.A.

@class ASIPithosRequest;
@class ASIPithos;
@class ASIPithosContainerRequest;
@class ASIPithosObjectRequest;
@class ASIPithosObject;

@interface PithosUtilities : NSObject

+ (ASIPithosObjectRequest *)objectDataRequestWithPithos:(ASIPithos *)pithos 
                                          containerName:(NSString *)containerName 
                                             objectName:(NSString *)objectName 
                                                version:(NSString *)version 
                                            toDirectory:(NSString *)directoryPath 
                                        withNewFileName:(NSString *)newFileName 
                                          checkIfExists:(BOOL)ifExists 
                                         sharingAccount:(NSString *)sharingAccount;
+ (NSArray *)objectDataRequestsForSubdirWithPithos:(ASIPithos *)pithos 
                                     containerName:(NSString *)containerName 
                                        objectName:(NSString *)objectName 
                                       toDirectory:(NSString *)directoryPath 
                                     checkIfExists:(BOOL)ifExists 
                                    sharingAccount:(NSString *)sharingAccount;

+ (ASIPithosObjectRequest *)objectBlockDataRequestWithPithos:(ASIPithos *)pithos 
                                               containerName:(NSString *)containerName 
                                                      object:(ASIPithosObject *)object 
                                                  blockIndex:(NSUInteger)blockIndex 
                                                   blockSize:(NSUInteger)blockSize;
+ (NSIndexSet *)missingBlocksForFile:(NSString *)filePath
                           blockSize:(NSUInteger)blockSize 
                           blockHash:(NSString *)blockHash 
                          withHashes:(NSArray *)hashes;

+ (ASIPithosObjectRequest *)writeObjectDataRequestWithPithos:(ASIPithos *)pithos 
                                               containerName:(NSString *)containerName
                                                  objectName:(NSString *)objectName
                                                 contentType:(NSString *)contentType 
                                                   blockSize:(NSUInteger)blockSize 
                                                   blockHash:(NSString *)blockHash 
                                                     forFile:(NSString *)filePath 
                                               checkIfExists:(BOOL)ifExists 
                                                      hashes:(NSArray **)hashes 
                                              sharingAccount:(NSString *)sharingAccount;
+ (NSIndexSet *)missingBlocksForHashes:(NSArray *)hashes withMissingHashes:(NSArray *)missingHashes;
+ (ASIPithosContainerRequest *)updateContainerDataRequestWithPithos:(ASIPithos *)pithos 
                                                      containerName:(NSString *)containerName 
                                                          blockSize:(NSUInteger)blockSize 
                                                            forFile:(NSString *)filePath 
                                                             hashes:(NSArray *)hashes 
                                                      missingHashes:(NSArray *)missingHashes 
                                                     sharingAccount:(NSString *)sharingAccount;
+ (ASIPithosContainerRequest *)updateContainerDataRequestWithPithos:(ASIPithos *)pithos 
                                                      containerName:(NSString *)containerName 
                                                          blockSize:(NSUInteger)blockSize 
                                                            forFile:(NSString *)filePath 
                                                  missingBlockIndex:(NSUInteger)missingBlockIndex 
                                                     sharingAccount:(NSString *)sharingAccount;
+ (NSArray *)writeObjectDataRequestsWithPithos:(ASIPithos *)pithos 
                                 containerName:(NSString *)containerName
                                    objectName:(NSString *)objectName
                                     blockSize:(NSUInteger)blockSize 
                                     blockHash:(NSString *)blockHash 
                                  forDirectory:(NSString *)directoryPath 
                                 checkIfExists:(BOOL)ifExists 
                                   objectNames:(NSMutableArray **)objectNames
                                  contentTypes:(NSMutableArray **)contentTypes
                                     filePaths:(NSMutableArray **)filePaths 
                                  hashesArrays:(NSMutableArray **)hashesArrays 
                       directoryObjectRequests:(NSMutableArray **) directoryObjectRequests 
                                sharingAccount:(NSString *)sharingAccount;

+ (NSArray *)deleteObjectRequestsForSubdirWithPithos:(ASIPithos *)pithos 
                                       containerName:(NSString *)containerName 
                                          objectName:(NSString *)objectName;

+ (ASIPithosObjectRequest *)cpyObjectRequestWithPithos:(ASIPithos *)pithos 
                                          containerName:(NSString *)containerName 
                                             objectName:(NSString *)objectName 
                               destinationContainerName:(NSString *)destinationContainerName 
                                  destinationObjectName:(NSString *)destinationObjectName 
                                          checkIfExists:(BOOL)ifExists 
                                         sharingAccount:(NSString *)sharingAccount;
+ (NSArray *)cpyObjectRequestsForSubdirWithPithos:(ASIPithos *)pithos 
                                     containerName:(NSString *)containerName 
                                        objectName:(NSString *)objectName 
                          destinationContainerName:(NSString *)destinationContainerName 
                             destinationObjectName:(NSString *)destinationObjectName 
                                     checkIfExists:(BOOL)ifExists 
                                    sharingAccount:(NSString *)sharingAccount;

+ (ASIPithosObjectRequest *)moveObjectRequestWithPithos:(ASIPithos *)pithos 
                                          containerName:(NSString *)containerName 
                                             objectName:(NSString *)objectName 
                               destinationContainerName:(NSString *)destinationContainerName 
                                  destinationObjectName:(NSString *)destinationObjectName 
                                          checkIfExists:(BOOL)ifExists;
+ (NSArray *)moveObjectRequestsForSubdirWithPithos:(ASIPithos *)pithos 
                                     containerName:(NSString *)containerName 
                                        objectName:(NSString *)objectName 
                          destinationContainerName:(NSString *)destinationContainerName 
                             destinationObjectName:(NSString *)destinationObjectName 
                                     checkIfExists:(BOOL)ifExists;

+ (NSUInteger)bytesOfFile:(NSString *)filePath;
+ (NSString *)contentTypeOfFile:(NSString *)filePath error:(NSError **)error;
+ (BOOL)safeCreateDirectory:(NSString *)directoryPath error:(NSError **)error;
+ (void)removeContentsAtPath:(NSString *)dirPath andDirectory:(BOOL)removeDirectory;
+ (void)removeContentsAtPath:(NSString *)dirPath;
+ (BOOL)isContentTypeDirectory:(NSString *)contentType;
+ (BOOL)objectExistsAtPithos:(ASIPithos *)pithos containerName:(NSString *)containerName objectName:(NSString *)objectName 
                       error:(NSError **)error isDirectory:(BOOL *)isDirectory sharingAccount:(NSString *)sharingAccount;
+ (BOOL)proceedIfObjectExistsAtPithos:(ASIPithos *)pithos containerName:(NSString *)containerName objectName:(NSString *)objectName 
                       sharingAccount:(NSString *)sharingAccount;
+ (NSArray *)objectsWithPithos:(ASIPithos *)pithos containerName:(NSString *)containerName objectNamePrefix:(NSString *)objectNamePrefix 
                     delimiter:(NSString *)delimiter sharingAccount:(NSString *)sharingAccount;
+ (NSArray *)objectsForSubdirWithPithos:(ASIPithos *)pithos containerName:(NSString *)containerName objectName:(NSString *)objectName 
                              delimiter:(NSString *)delimiter sharingAccount:(NSString *)sharingAccount;
+ (NSString *)safeObjectNameForPithos:(ASIPithos *)pithos containerName:(NSString *)containerName objectName:(NSString *)objectName;
+ (NSString *)safeSubdirNameForPithos:(ASIPithos *)pithos containerName:(NSString *)containerName subdirName:(NSString *)subdirName;

+ (void)httpRequestErrorAlertWithRequest:(ASIPithosRequest *)request;
+ (void)unexpectedResponseStatusAlertWithRequest:(ASIPithosRequest *)request;
+ (void)httpAuthenticationError;
+ (void)fileActionFailedAlertWithTitle:(NSString *)title message:(NSString *)message error:(NSError *)error;

+ (ASIPithosRequest *)prepareRequest:(ASIPithosRequest *)request priority:(NSOperationQueuePriority)priority;
+ (ASIPithosRequest *)prepareRequest:(ASIPithosRequest *)request;
+ (ASIPithosRequest *)copyRequest:(ASIPithosRequest *)request;
+ (void)startAndWaitForRequest:(ASIPithosRequest *)request;

@end
